/********************************************************************/
/*                                                                  */
/*  Licensed Materials - Property of IBM.                           */
/*  5724-T42                                                        */
/*  5724-T43                                                        */
/*  Copyright IBM Corp. 2008.                                       */
/*  All Rights Reserved.                                            */
/*  US Government Users Restricted Rights -                         */
/*  Use, duplication or disclosure restricted by                    */
/*  GSA ADP Schedule Contract with IBM Corp.                        */
/*                                                                  */
/*  This file contains sample code.  You may copy, modify, and      */
/*  distribute these samples, or their modifications, in any        */
/*  form, internally or as part of Your application or related      */
/*  documentation. These samples have not been tested under all     */
/*  conditions and are provided to You by IBM without obligation    */
/*  of support of any kind. IBM PROVIDES THESE SAMPLES "AS IS"      */
/*  SUBJECT TO ANY STATUTORY WARRANTIES THAT CANNOT BE EXCLUDED.    */
/*  IBM MAKES NO WARRANTIES OR CONDITIONS, EITHER EXPRESS OR        */
/*  IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES   */
/*  OR CONDITIONS OF MERCHANTABILITY, FITNESS FOR A PARTICULAR      */
/*  PURPOSE, AND NON-INFRINGEMENT REGARDING THESE SAMPLES OR        */
/*  TECHNICAL SUPPORT, IF ANY. You will indemnify IBM or third      */
/*  parties that provide IBM products ("Third Parties") from and    */
/*  against any third party claim arising out of the use,           */
/*  modification or distribution of these samples with Your         */
/*  application. You may not use the same path name as the          */
/*  original files/modules. You must not alter or delete any        */
/*  copyright information in the Samples.                           */
/*                                                                  */
/********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/***************************************************
 Explicit Finite difference scheme for computing 
 European Option (with dividend)
 Input Parameters:
 Strike Value K 
 Time T 
 Current Stock Price S 
 volality sigma sigma (in decimal - not in %)
 rate of return r (in decimal - not in %)
 dividend div (in decimal - not in %) 
 Number of Time Steps N_Time_Steps 
 Number of Stock Steps N_S_Divisions  
 dx 
***************************************************/

static inline double max(double a, double b) {
	return a > b ? a : b;
}


double explicit_EO_call (double K,
			 double T, double S, double sigma,
			 double r, double div, int N_Time_Steps,
			 int N_S_Divisions, double dx) {

	double HALF = (double)0.5;
	double ONE  = (double)1.0;
	double ZERO = (double)0.0;
	double dt, ds;		/* dt is the time step, ds is the grid points on S */
	double nu, edx ;
	int Total_N_S_Div;
	double pu, pm, pd;	/* Coefficients for computing new value at (i,j) */
	int i, j, k; 		/* General loop integers */

	/* We need 3 arrays (vectors) - one for holding S values, one for holding old T values
	   and one for holding New T values.  New T gets copied to Old T as we time-step */

	double S_Values[2048] __attribute ((aligned (16)));
	double F_Old[2048]    __attribute ((aligned (16)));
	double F_New[2048]    __attribute ((aligned (16)));

	double sigma_sqr = sigma * sigma ;
	double zeta      = sigma/dx;
	double zeta_sqr  = zeta * zeta ;

	dt  = T/(double) N_Time_Steps;
	nu  =  (r-div-(HALF*sigma_sqr));
	edx = exp(dx);

	pu = HALF * dt * (zeta_sqr + (nu/dx));	/* Coefficient for j+1 term */
	pm = ONE - dt*zeta_sqr - r*dt;   	/* Coefficient for j term */
	pd = HALF * dt * (zeta_sqr - (nu/dx));	/* Coefficient for j-1 term */

	/* Malloc the arrays we created to problem size i.e. N_S_Divisions */
	Total_N_S_Div = (2*N_S_Divisions)+1;

	/* Fill the S_Values */
	S_Values[0] = S*exp(-N_Time_Steps*dx);
	for(i=1; i<Total_N_S_Div; i++) {
		S_Values[i] = S_Values[i-1]*edx;
	}

	/* Fill in the (Right) Boundary values - Corrsponds to values at time T*/
	for(i=0; i< Total_N_S_Div; i++) {
		F_Old[i] = (double) max(ZERO,S_Values[i]-K);
	}
        
	/* Step through T from Right to Left and compute the values */
	double *newPtr = F_New, *oldPtr = F_Old, *tPtr;                 
	for(j=N_Time_Steps; j>0; --j) {
		F_New[0] = F_Old[0];  /* Boundary condition */
		for (i=1; i<(Total_N_S_Div-1); i++) {
			F_New[i] = (pu*F_Old[i+1])+(pm*F_Old[i])+(pd*F_Old[i-1]);
		}
		/* Boundary value */
		F_New[Total_N_S_Div-1] = F_New[Total_N_S_Div-2] 
					+ S_Values[Total_N_S_Div-1] 
					- S_Values[Total_N_S_Div -2];

		/* Copy the New Vector into old Vector */
		for(k=0; k<Total_N_S_Div; k++) {
			F_Old[k] = F_New[k];
		}
	}

	return F_New[N_S_Divisions];
}
